@@ -1,8 +1,12 @@
+# (c) 2005 Ian Bicking and contributors; written for Paste (http://pythonpaste.org)
+# Licensed under the MIT license: http://www.opensource.org/licenses/mit-license.php
 import os
 import re
+import sys
 import urllib
 from ConfigParser import ConfigParser
 import pkg_resources
+from paste.deploy.util.fixtypeerror import fix_call
 
 __all__ = ['loadapp', 'loadserver', 'loadfilter', 'appconfig']
 
@@ -74,7 +78,8 @@ class _ObjectType(object):
 
     def invoke(self, context):
         assert context.protocol in _flatten(self.egg_protocols)
-        return context.object(context.global_conf, **context.local_conf)
+        return fix_call(context.object,
+                        context.global_conf, **context.local_conf)
 
 class _App(_ObjectType):
 
@@ -87,10 +92,11 @@ class _App(_ObjectType):
     def invoke(self, context):
         if context.protocol in ('paste.composit_factory',
                                 'paste.composite_factory'):
-            return context.object(context.loader, context.global_conf,
-                                  **context.local_conf)
+            return fix_call(context.object,
+                            context.loader, context.global_conf,
+                            **context.local_conf)
         elif context.protocol == 'paste.app_factory':
-            return context.object(context.global_conf, **context.local_conf)
+            return fix_call(context.object, context.global_conf, **context.local_conf)
         else:
             assert 0, "Protocol %r unknown" % context.protocol
 
@@ -103,12 +109,14 @@ class _Filter(_ObjectType):
 
     def invoke(self, context):
         if context.protocol == 'paste.filter_factory':
-            return context.object(context.global_conf, **context.local_conf)
+            return fix_call(context.object,
+                            context.global_conf, **context.local_conf)
         elif context.protocol == 'paste.filter_app_factory':
             def filter_wrapper(wsgi_app):
                 # This should be an object, so it has a nicer __repr__
-                return context.object(wsgi_app, context.global_conf,
-                                      **context.local_conf)
+                return fix_call(context.object,
+                                wsgi_app, context.global_conf,
+                                **context.local_conf)
             return filter_wrapper
         else:
             assert 0, "Protocol %r unknown" % context.protocol
@@ -122,12 +130,14 @@ class _Server(_ObjectType):
 
     def invoke(self, context):
         if context.protocol == 'paste.server_factory':
-            return context.object(context.global_conf, **context.local_conf)
+            return fix_call(context.object,
+                            context.global_conf, **context.local_conf)
         elif context.protocol == 'paste.server_runner':
             def server_wrapper(wsgi_app):
                 # This should be an object, so it has a nicer __repr__
-                return context.object(wsgi_app, context.global_conf,
-                                      **context.local_conf)
+                return fix_call(context.object,
+                                wsgi_app, context.global_conf,
+                                **context.local_conf)
             return server_wrapper
         else:
             assert 0, "Protocol %r unknown" % context.protocol
@@ -213,6 +223,8 @@ def loadcontext(object_type, uri, name=None, relative_to=None,
             uri = uri.split('#', 1)[0]
     if name is None:
         name = 'main'
+    if ':' not in uri:
+        raise LookupError("URI has no scheme: %r" % uri)
     scheme, path = uri.split(':', 1)
     scheme = scheme.lower()
     if scheme not in _loaders:
@@ -228,7 +240,15 @@ def _loadconfig(object_type, uri, path, name, relative_to,
                 global_conf):
     # De-Windowsify the paths:
     path = path.replace('\\', '/')
-    if not path.startswith('/'):
+    absolute_path = True
+    if sys.platform == 'win32':
+        _absolute_re = re.compile(r'^[a-zA-Z]:')
+        if not _absolute_re.search(path):
+            absolute_path = False
+    else:
+        if not path.startswith('/'):
+            absolute_path = False
+    if not absolute_path:
         if not relative_to:
             raise ValueError(
                 "Cannot resolve relative uri %r; no context keyword "
@@ -242,6 +262,8 @@ def _loadconfig(object_type, uri, path, name, relative_to,
         path = path[2:]
     path = urllib.unquote(path)
     loader = ConfigLoader(path)
+    if global_conf:
+        loader.update_defaults(global_conf, overwrite=False)
     return loader.get_context(object_type, name, global_conf)
 
 _loaders['config'] = _loadconfig
@@ -310,6 +332,12 @@ class ConfigLoader(_Loader):
         self.parser._defaults.setdefault(
             '__file__', os.path.abspath(filename))
 
+    def update_defaults(self, new_defaults, overwrite=True):
+        for key, value in new_defaults.items():
+            if not overwrite and key in self.parser._defaults:
+                continue
+            self.parser._defaults[key] = value
+
     def get_context(self, object_type, name=None, global_conf=None):
         if self.absolute_name(name):
             return loadcontext(object_type, name,
